# Copyright(c) 2024 Association of Universities for Research in Astronomy, Inc.

procedure nfapsum (inimages)

# Integrate flux over IFU slices/exposures to get a single 1D spectrum

char    inimages    {prompt="Input images"}
char    outspectrum {"",prompt="Output image with 1D spectrum"}
char    outprefix   {"i",prompt="Prefix for output image"}
char    extrange    {"1-", prompt="Range of SCI extver to include in sum"}
bool    fl_vardq    {yes, prompt = "Create variance and data quality frames?"}
char    logfile     {"", prompt = "Logfile"}
bool    verbose     {yes, prompt = "Verbose"}
int     status      {0, prompt = "Exit status (0=good)"}
struct *scanfile1   {"", prompt = "Internal use only"}
struct *scanfile2   {"", prompt = "Internal use only"}

begin

    char    l_inimages = ""
    char    l_outspectrum = ""
    char    l_outprefix = ""
    char    l_extrange = ""
    bool    l_fl_vardq

    char    l_logfile = ""
    bool    l_verbose

    char    l_sci_ext = ""
    char    l_var_ext = ""
    char    l_dq_ext = ""
    char    l_key_dispaxis = ""

    int     corners = 8       # not usual unillum flag, but what gnirs pkg uses
    int     incomplete = 128  # summed over good pix but missing counts
    int     bitmask = 121     # bad pix to exclude from sum (not just flag)

    int     junk, dispaxis, nfiles, nsci, extver, ndim, npix, tnpix
    int     nimages, ntot
    char    badhdr, tmpinimages, tmpout, refimg, refext, img
    char    phu, keyfound, text, tmpscilist, sciext, varext, dqext
    char    tmpsci, tmpvar, tmpbad, tmpcount
    char    tmpsci1d, tmpvar1d, tmpbad1d, tmpcount1d, tmpdq1d, tmpwcs1d
    char    tmpoutsci, tmpoutvar, tmpoutdq, tmpexpr, dqexpr
    bool    firstimg, firstext, havedq, incvar, incdq
    struct  sdate

    status = 1
    cache ("gnirs", "nsheaders", "gemextn", "keypar", "gemdate")

    junk = fscan (inimages, l_inimages)
    junk = fscan (outspectrum, l_outspectrum)
    junk = fscan (outprefix, l_outprefix)
    junk = fscan (extrange, l_extrange)
    l_fl_vardq = fl_vardq

    junk = fscan (logfile, l_logfile)
    l_verbose = verbose

    badhdr = ""
    junk = fscan (nsheaders.sci_ext, l_sci_ext)
    if ("" == l_sci_ext) badhdr = badhdr + " sci_ext"
    junk = fscan (nsheaders.var_ext, l_var_ext)
    if ("" == l_var_ext) badhdr = badhdr + " var_ext"
    junk = fscan (nsheaders.dq_ext, l_dq_ext)
    if ("" == l_dq_ext) badhdr = badhdr + " dq_ext"
    junk = fscan (nsheaders.key_dispaxis, l_key_dispaxis)
    if ("" == l_key_dispaxis) badhdr = badhdr + " key_dispaxis"
    # junk = fscan (nsheaders.key_instrument, l_key_instrument)
    # if ("" == l_key_instrument) badhdr = badhdr + " key_instrument"

    if ("" == l_logfile) {
        junk = fscan (gnirs.logfile, l_logfile)
        if ("" == l_logfile) {
            l_logfile = "gnirs.log"
            printlog ("WARNING - NFAPSUM: Both nfapsum.logfile and \
                gnirs.logfile are", l_logfile, verbose+) 
            printlog ("                     undefined. Using " \
                // l_logfile, l_logfile, verbose+)
        }
    }

    date | scan (sdate) 
    printlog ("----------------------------------------------------------\
        ---------------------", l_logfile, l_verbose) 
    printlog ("NFAPSUM -- " // sdate, l_logfile, l_verbose) 
    printlog (" ", l_logfile, l_verbose)

    printlog ("inimages    = "//l_inimages, l_logfile, l_verbose)
    printlog ("outspectrum = "//l_outspectrum, l_logfile, l_verbose)
    printlog ("outprefix   = "//l_outprefix, l_logfile, l_verbose)
    printlog ("extrange    = "//l_extrange, l_logfile, l_verbose)
    printlog ("fl_vardq    = "//l_fl_vardq, l_logfile, l_verbose)

    printlog ("", l_logfile, l_verbose)

    if ("" != badhdr) {
        printlog ("ERROR - NFAPSUM: Parameter(s) missing from \
            nsheaders: " // badhdr, l_logfile, verbose+) 
        goto clean
    }

    tmpinimages = mktemp ("tmpinimages")
    tmpout = mktemp ("tmpout")
    tmpscilist = mktemp ("tmpscilist")
    tmpsci = mktemp ("tmpsci")
    tmpvar = mktemp ("tmpvar")
    tmpbad = mktemp ("tmpbad")
    tmpcount = mktemp ("tmpcount")
    tmpsci1d = mktemp ("tmpsci1d")
    tmpvar1d = mktemp ("tmpvar1d")
    tmpbad1d = mktemp ("tmpbad1d")
    tmpcount1d = mktemp ("tmpcount1d")
    tmpdq1d = mktemp ("tmpdq1d")
    tmpwcs1d = mktemp ("tmpwcs1d")
    tmpoutsci = mktemp ("tmpoutsci")
    tmpoutvar = mktemp ("tmpoutvar")
    tmpoutdq = mktemp ("tmpoutdq")

    gemextn (l_inimages, check="exists,mef", process="none", index="",
        extname="", extversion="", ikparams="", omit="extension,kernel",
        replace="", outfile=tmpinimages, logfile="", glogpars="",
        verbose=l_verbose)
    if (0 < gemextn.fail_count) {
        printlog ("ERROR - NFAPSUM: Bad or missing file in inimages.",
            l_logfile, verbose+) 
        goto clean
    }
    nfiles = gemextn.count

    gemextn (l_outspectrum, proc="none", check="absent", index="",
        extname="", extver="", ikparams="", omit="", replace="",
        outfile="dev$null", logfile="", glogpars="", verbose-)
    if (0 < gemextn.fail_count) {
        printlog ("ERROR - NFAPSUM: Output image already exists.",
            l_logfile, verbose+) 
        goto clean
    }
    head (tmpinimages, nlines=1) | scan (refimg)
    if (0 == gemextn.count) {
      l_outspectrum = l_outprefix//refimg
      gemextn (l_outspectrum, proc="none", check="absent", index="",
          extname="", extver="", ikparams="", omit="", replace="",
          outfile="dev$null", logfile="", glogpars="", verbose-)
          if (0 < gemextn.fail_count || 1 != gemextn.count) {
              printlog ("ERROR - NFAPSUM: Bad output files.", \
                  l_logfile, verbose+) 
              goto clean
          }
    }

    incvar = l_fl_vardq
    incdq = l_fl_vardq

    # Loop over input images:
    nimages = 0
    ntot = 0
    firstimg = yes
    firstext = yes
    scanfile1 = tmpinimages
    while (fscan (scanfile1, img) != EOF) {

        # Check the input file and determine properties like dimensionality:

        phu = img // "[0]"

        keyfound = ""
        hselect(phu, "*PREPAR*", yes) | scan(keyfound)
        if (keyfound == "") {
            printlog ("ERROR - NFAPSUM: Image " // img // " not \
                PREPAREd.", l_logfile, verbose+)
            goto clean
        }

        gemextn (img, check="", process="expand", index="", extname=l_sci_ext,
            extversion=l_extrange, ikparams="", omit="params,section",
            replace="", outfile=tmpscilist, logfile="", glogpars="",
            verbose=l_verbose)
        nsci = gemext.count

        refext = img//"["//l_sci_ext//",1]"

        hselect (refext, "i_naxis", yes) | scan (ndim)

        dispaxis = INDEF
        hselect (refext, l_key_dispaxis, yes) | scan (dispaxis)
        if (isindef(dispaxis)) {
            if (ndim < 2) dispaxis = 1
            else dispaxis = 2  # default for GNIRS
        }

        # Create a working output image & determine its dimension.
        # There's no point retaining the MDF when integrating over slices.
        if (firstimg) {
          imcopy (phu, tmpout, verbose-)
          hselect (refext, "i_naxis"//dispaxis, yes) | scan (npix)
          # hselect (refext, "CTYPE"//dispaxis//",CRPIX"//dispaxis//",CRVAL" \
          #     //dispaxis//",CD"//dispaxis//"_"//dispaxis//",CUNIT"//dispaxis,
          #     yes) | scan (ctype, crpix, crval, cdelt, cunit)
          firstimg = no
        }

        # Loop over input extensions:
        scanfile2 = tmpscilist
        while (fscan (scanfile2, sciext) != EOF) {

            # Re-parse the version from the SCI extension, since it appears
            # that gemextn can't simply list version numbers:
            text = substr (sciext, strldx ("[", sciext) + 1,
                                   strldx ("]", sciext) - 1)
            extver = int (substr (text, stridx (",", text) + 1, strlen(text)))

            varext = img // "[" // l_var_ext // "," // extver // "]"
            dqext = img // "[" // l_dq_ext // "," // extver // "]"

            # Check whether we have VAR and complain if it's missing when
            # propagation is enabled:
            if (no == imaccess(varext)) {
                incvar = no
                if (l_fl_vardq)
                    printlog ("WARNING - NFAPSUM: Missing VAR for " \
                        // l_sci_ext // "," // extver, l_logfile, verbose+)
            }

            # If DQ is missing, omit masking but make sure the user knows,
            # since we probably want to exclude bad/unilluminated pixels even
            # when we're not propagating DQ to the output.
            if (no == imaccess(dqext)) {
                havedq = no
                incdq = no
                printlog ("WARNING - NFAPSUM: Missing DQ for " \
                    // l_sci_ext // "," // extver, l_logfile, verbose+)
            } else havedq = yes

            # The lengths of all the input spectra must match:
            hselect (sciext, "i_naxis"//dispaxis, yes) | scan (tnpix)
            if (tnpix != npix) {
                printlog ("ERROR - NFAPSUM: dimension mismatch along \
                    dispersion axes of inputs.", l_logfile, verbose+)
                goto clean
            }

            # Integrate input over its spatial axis, rejecting bad pixels:
            if (havedq) {
                tmpexpr = "c & "//bitmask//" ? 0 : "
                dqexpr = "c & (("//bitmask//"^"//corners//")|"//incomplete//")\
                          ? 1 : 0*c"
            } else {
                tmpexpr = ""
            }

            # Convert bad pixels to 0 (& count them in the "DQ", only within
            # the illuminated area):
            imexpr (tmpexpr//"a", tmpsci, a=sciext, b=varext, c=dqext,
                    refim="a", outtype="ref", verbose-)
            if (incvar)
                imexpr (tmpexpr//"b", tmpvar, a=sciext, b=varext, c=dqext,
                    refim="b", outtype="ref", verbose-)
            if (incdq)
                imexpr (dqexpr, tmpbad, a=sciext, b=varext, c=dqext,
                    refim="c", outtype="ref", verbose-)

            # Flag which pixels were included (the input images only have to
            # exist if used in the expression, which has to include "a" in
            # order to use it as the reference and guarantee the right dims):
            imexpr (tmpexpr//"0*a + 1", tmpcount, a=sciext, b=varext, c=dqext,
                refim="a", outtype="int", verbose-)

            # Integrate over spatial axis, if applicable, to get a 1D spec:
            if (ndim == 1) {
                imcopy (tmpsci, tmpsci1d, verbose-)
                imcopy (tmpcount, tmpcount1d, verbose-)
                if (incvar)
                    imcopy (tmpvar, tmpvar1d, verbose-)
                if (incdq)
                    imcopy (tmpbad, tmpbad1d, verbose-)
            } else {
                improject (tmpsci, tmpsci1d, projaxis=3-dispaxis, average-,
                    highcut=0., lowcut=0., pixtype="real", verbose-)
                improject (tmpcount, tmpcount1d, projaxis=3-dispaxis, average-,
                           highcut=0, lowcut=0, pixtype="int", verbose-)
                # Simplify the resulting 2D WCS to 1D:
                scopy (tmpsci1d, tmpwcs1d, w1=INDEF, w2=INDEF, apertures="",
                    bands="", beams="", apmodulus=0, format="multispec",
                    renumber-, offset=0, clobber+, merge-, rebin-, verbose-)
                wcscopy (tmpsci1d, tmpwcs1d, verbose-)
                if (incvar) {
                    improject (tmpvar, tmpvar1d, projaxis=3-dispaxis, average-,
                        highcut=0., lowcut=0., pixtype="real", verbose-)
                    wcscopy (tmpvar1d, tmpwcs1d, verbose-)
                }
                if (incdq) {
                    improject (tmpbad, tmpbad1d, projaxis=3-dispaxis, average-,
                        highcut=0., lowcut=0., pixtype="int", verbose-)
                    wcscopy (tmpbad1d, tmpwcs1d, verbose-)
                }
            }
            if (incdq) {
                imexpr ("(a > 0 ? "//incomplete//" : 0) | \
                         (b > 0 ? 0 : "//corners//")",
                    tmpdq1d, a=tmpbad1d, b=tmpcount1d, refim="a",
                    outtype="ushort", verbose-)
            }

            # Accumulate each 1D spectrum & its mask into the respective sums:
            if (firstext) {
                imcopy (tmpsci1d, tmpoutsci, verbose-)
                if (incvar)
                    imcopy (tmpvar1d, tmpoutvar, verbose-)
                if (incdq)
                    imcopy (tmpdq1d, tmpoutdq, verbose-)
                firstext = no
            } else {
                imarith (tmpoutsci, "+", tmpsci1d, tmpoutsci, hparams="",
                    pixtype="", calctype="", verbose-, noact-)
                if (incvar) {
                    imarith (tmpoutvar, "+", tmpvar1d, tmpoutvar, hparams="",
                        pixtype="", calctype="", verbose-, noact-)
                }
                if (incdq) {
                    imdelete (tmpbad1d, verify-, >& "dev$null")
                    imexpr ("c > 0 ? a & ~"//corners//" : a | b", tmpbad1d,
                        a=tmpoutdq, b=tmpdq1d, c=tmpcount1d, refim="a",
                        outtype="ref", verbose-)
                    imdelete (tmpoutdq, verify-, >& "dev$null")
                    imcopy (tmpbad1d, tmpoutdq, verbose-)
                }
            }
            ntot += 1

            imdelete (tmpsci//","//tmpvar//","//tmpbad, verify-, >& "dev$null")
            imdelete (tmpsci1d//","//tmpvar1d, verify-, >& "dev$null")
            imdelete (tmpbad1d//","//tmpdq1d, verify-, >& "dev$null")
            imdelete (tmpcount//","//tmpcount1d, verify-, >& "dev$null")
            imdelete (tmpwcs1d, verify-, >& "dev$null")

        }  # (end loop over extensions)

        nimages += 1

        delete (tmpscilist, verify-, >& "dev$null")

    }  # (end loop over files)

    # Finish assembling the output. We can't scale the output according to how
    # many bad pixels were excluded, since the flux is usually concentrated in
    # a few columns in normal seeing, but we can try to indicate with DQ=1
    # which pixels come from incomplete inputs (not considering "corners").
    # imstat (tmpoutmsk, fields="max", lower=INDEF, upper=INDEF, nclip=0,
    #     format-, cache-) | scan (maxpix)
    imexpr ("a", tmpout//"["//l_sci_ext//",1,append]",
        a=tmpoutsci, b=tmpoutvar, c=tmpoutdq, refim="a", outtype="ref",
        verbose-)
    if (incvar) {
        imexpr ("b", tmpout//"["//l_var_ext//",1,append]",
            a=tmpoutsci, b=tmpoutvar, c=tmpoutdq, refim="b", outtype="ref",
            verbose-)
    }
    if (incdq) {
        imexpr ("c", tmpout//"["//l_dq_ext//",1,append]",
            a=tmpoutsci, b=tmpoutvar, c=tmpoutdq, refim="c", outtype="ref",
            verbose-)
    }

    printlog ("NFAPSUM: Combined "//ntot//" extensions from "//nimages \
        //" image(s)", l_logfile, l_verbose)

    # Task was successful if we've reached this point:
    imrename(tmpout, l_outspectrum, verbose-)
    status = 0

clean:

    delete (tmpinimages, verify-, >& "dev$null")
    delete (tmpscilist, verify-, >& "dev$null")
    imdelete (tmpsci//","//tmpvar//","//tmpbad, verify-, >& "dev$null")
    imdelete (tmpsci1d//","//tmpvar1d//","//tmpbad1d, verify-, >& "dev$null")
    imdelete (tmpoutsci//","//tmpoutvar//","//tmpoutdq, verify-, >& "dev$null")
    imdelete (tmpcount//","//tmpcount1d//","//tmpwcs1d, verify-, >& "dev$null")
    imdelete (tmpout, verify-, >& "dev$null")

    printlog (" ", l_logfile, l_verbose) 
    if (status == 0) {
        printlog ("NFAPSUM exit status: good.", l_logfile, l_verbose) 
    }
    printlog ("-------------------------------------------------------\
        -----------------------", l_logfile, l_verbose) 

    scanfile1 = ""
    scanfile2 = ""

end
