# Copyright(c) 2001-2024 Association of Universities for Research in Astronomy, Inc.

procedure nfnormslices (ifuflats, output) 

# Correct the slice-to-slice variations for GNIRS IFU (based on nsslitfunction)

string  ifuflats    {prompt = "Flat-fielded, rectified, spatially-flat inputs"}
string  output      {prompt = "Output normalized flat (MEF)"}
string  flat        {"", prompt = "Input processed lamp flat"}
string  title       {"", prompt = "Title for output SCI plane"}
string  combine     {"median", prompt = "Combination operation"}
string  reject      {"none", prompt = "Rejection algorithm"}
#string  scale       {"median", prompt = "Image scaling"}
#string  zero        {"none", prompt = "Image zeropoint offset"}
#string  weight      {"none", prompt = "Image weights"}
string  statsec     {"default", prompt = "Statistics section"}
real    lthreshold  {INDEF, prompt = "Lower threshold"}
real    hthreshold  {INDEF, prompt = "Upper threshold"}
int     nlow        {1, min=1, prompt = "minmax: Number of low pixels to reject"}
int     nhigh       {1, min=1, prompt = "minmax: Number of high pixels to reject"}
int     nkeep       {0, min=0, prompt = "Minimum to keep or maximum to reject"}  
bool    mclip       {yes, prompt = "Use median in sigma clipping algorithms?"}
real    lsigma      {3., prompt = "Lower sigma clipping factor"}
real    hsigma      {3., prompt = "Upper sigma clipping factor"}
real    ron         {0.0, min = 0., prompt = "Readout noise rms in electrons"}
real    gain        {1.0, min = 0.00001, prompt = "Gain in e-/ADU"}
string  snoise      {"0.0", prompt = "ccdclip: Sensitivity noise (electrons)"}
#real    sigscale    {0.1, prompt = "Tolerance for sigma clipping scaling correction"}
real    pclip       {-0.5, prompt = "pclip: Percentile clipping parameter"}
real    grow        {0.0, prompt = "Radius (pixels) for neighbor rejection"}
string  logfile     {"", prompt = "Logfile name"}
bool    verbose     {yes, prompt = "Verbose output?"}
int     status      {0, prompt = "Exit status (0=good)"}
struct  *scanfile   {"", prompt = "Internal use only"} 

begin

    string  l_ifuflats = ""
    string  l_flat = ""
    string  l_output = ""
    string  l_title = ""
    string  l_combine = ""
    string  l_reject = ""
    string  l_masktype = ""
    real    l_maskvalue
    string  l_scale = ""
    string  l_zero = ""
    string  l_weight = ""
    string  l_statsec = ""
    real    l_lthreshold
    real    l_hthreshold
    int     l_nlow
    int     l_nhigh
    int     l_nkeep
    bool    l_mclip
    real    l_lsigma
    real    l_hsigma
    real    l_ron
    real    l_gain
    string  l_snoise = ""
    real    l_sigscale
    real    l_pclip
    real    l_grow
    string  l_logfile = ""
    bool    l_verbose

    string  l_sci_ext
    string  l_var_ext
    string  l_dq_ext
    string  l_key_ron
    string  l_key_gain
    string  l_key_dispaxis
    string  l_key_exptime

    string  filelist, scilist, dqlist, specsec1, phulist
    string  img, varimg, dqimg, dqsum, dqsumold, suf, illumvar, tmplog, phu
    string  tmpdq, combout, combsig, combdq, illum, badhdr
    string  tmpimg
    real    norm, stddev
    int     i, k, n, j, l, idum, nbad, len, junk, nsci, iextn, nextn
    int     iver, nver, nx, ny, dispaxis, tnx, tny
    string  keyfound
    struct  sdate, line
    bool    debug, ok, havedqvar
    string  ssrclist, sdestlist, varlist, vsrclist, vdestlist, comblist
    string  dqsrclist, dqdestlist
    string  ssrc, sdest, vsrc, vdest, extn, skyextn, darkextn, flatextn
    string  comblist2, dqsrc, dqdest, sect
    real    mddata, ratio, correcn, oldcorrecn, oldmddata
    real    exptime, prevexp, sigma
    char    comboutptr, previmg, expr


    junk = fscan (  ifuflats, l_ifuflats)
    junk = fscan (  flat, l_flat)
    junk = fscan (  output, l_output)
    junk = fscan (  title, l_title)
    junk = fscan (  combine, l_combine)
    junk = fscan (  reject, l_reject)
    l_scale = "exposure"  # separate scaling of extensions not appropriate
    l_zero = "none"
    l_weight = "exposure"
    junk = fscan (  statsec, l_statsec)
    l_lthreshold =  lthreshold
    l_hthreshold =  hthreshold
    l_nlow =    nlow
    l_nhigh =   nhigh
    l_nkeep =   nkeep
    l_mclip =   mclip
    l_lsigma =  lsigma
    l_hsigma =  hsigma
    l_ron =     ron
    l_gain =    gain
    junk = fscan (  snoise, l_snoise)
    l_sigscale = 0.1
    l_pclip =       pclip
    l_grow =        grow
    junk = fscan (  logfile, l_logfile)
    l_verbose =     verbose

    badhdr = ""
    junk = fscan (  nsheaders.sci_ext, l_sci_ext)
    if ("" == l_sci_ext) badhdr = badhdr + " sci_ext"
    junk = fscan (  nsheaders.var_ext, l_var_ext)
    if ("" == l_var_ext) badhdr = badhdr + " var_ext"
    junk = fscan (  nsheaders.dq_ext, l_dq_ext)
    if ("" == l_dq_ext) badhdr = badhdr + " dq_ext"
    junk = fscan (  nsheaders.key_ron, l_key_ron)
    if ("" == l_key_ron) badhdr = badhdr + " key_ron"
    junk = fscan (  nsheaders.key_gain, l_key_gain)
    if ("" == l_key_gain) badhdr = badhdr + " key_gain"
    junk = fscan (  nsheaders.key_dispaxis, l_key_dispaxis)
    if ("" == l_key_dispaxis) badhdr = badhdr + " key_dispaxis"
    junk = fscan (  nsheaders.key_exptime, l_key_exptime)
    if ("" == l_key_exptime) badhdr = badhdr + " key_exptime"

    status = 1
    debug = no

    havedqvar = yes
    prevexp = INDEF

    cache ("keypar", "gnirs", "gemextn", "gemdate")

    filelist = mktemp ("tmpfilelist") 
    phulist = mktemp ("tmpphulist") 
    if (debug) print (phulist)
    scilist = mktemp ("tmpscilist")
    varlist = mktemp ("tmpvarlist")
    dqlist = mktemp ("tmpdqlist")
    combout = mktemp ("tmpcombout")
    ssrclist = mktemp ("tmpssrclist")
    sdestlist = mktemp ("tmpsdestlist")
    vsrclist = mktemp ("tmpvsrclist")
    vdestlist = mktemp ("tmpvdestlist")
    dqsrclist = mktemp ("tmpdqsrclist")
    dqdestlist = mktemp ("tmpdqdestlist")
    comblist = mktemp ("tmpcomblist")
    comblist2 = mktemp ("tmpcomblist2")
    tmpimg = mktemp ("tmpimg")


    # Start logging

    if ("" == l_logfile) {
        junk = fscan (gnirs.logfile, l_logfile) 
        if (l_logfile == "") {
            l_logfile = "gnirs.log"
            printlog ("WARNING - NFNORMSLICES: Both nfnormslices.logfile \
                and gnirs.logfile are empty.", l_logfile, verbose+) 
            printlog ("                   Using default file " \
                // l_logfile // ".", l_logfile, verbose+) 
        }
    }

    date | scan (sdate) 
    printlog ("-----------------------------------------------------\
        -------------------------", l_logfile, verbose = l_verbose) 
    printlog ("NFNORMSLICES -- " // sdate, l_logfile, l_verbose) 
    printlog (" ", l_logfile, l_verbose) 
    printlog ("ifuflats = " // l_ifuflats, l_logfile, l_verbose) 
    printlog ("output = " // l_output, l_logfile, l_verbose) 
    printlog ("title = " // l_title, l_logfile, l_verbose) 
    printlog ("sci_ext = " // l_sci_ext, l_logfile, l_verbose) 
    printlog ("var_ext = " // l_var_ext, l_logfile, l_verbose) 
    printlog ("dq_ext = " // l_dq_ext, l_logfile, l_verbose) 
    printlog (" ", l_logfile, l_verbose) 

    if ("" != badhdr) {
        printlog ("ERROR - NFNORMSLICES: Parameter(s) missing from \
            nsheaders: " // badhdr, l_logfile, verbose+) 
        goto clean
    }

    # Check/expand sky flats

    gemextn (l_ifuflats, check = "exists,mef", process = "none", \
        index = "", extname = "", extversion = "", ikparams = "", \
        omit = "", replace = "", outfile = filelist, logfile="",
        glogpars="", verbose=l_verbose)
    if (gemextn.fail_count != 0 || gemextn.count == 0) {
        printlog ("ERROR - NFNORMSLICES: Bad input images.", l_logfile, 
            verbose+) 
        goto clean
    }
    n = gemextn.count
    gemextn (l_ifuflats, check = "exists,mef", process = "expand", \
        index = "0", extname = "", extversion = "", ikparams = "", \
        omit = "", replace = "", outfile = phulist, logfile="",
        glogpars="", verbose=l_verbose)
    if (gemextn.fail_count != 0 || gemextn.count != n) {
        printlog ("ERROR - NFNORMSLICES: Bad or missing PHUs.", \
            l_logfile, verbose+) 
        goto clean
    }
    head (filelist, nlines=1) | scan (img)
    gemextn (img, check = "", process = "expand", index = "", \
        extname = l_sci_ext, extversion = "1-", ikparams = "", \
        omit = "", replace = "", outfile = "dev$null", \
        logfile="dev$null", glogpars="", verbose-)
    nsci = gemextn.count

    # fails if dispaxis is set for 1-D image, but code below will
    # handle that peculiar case for any likely value of dispaxis
    dispaxis = -1; nx = -1; ny = -1
    hselect (img//"["//l_sci_ext//",1]", "i_naxis1,i_naxis2,"//l_key_dispaxis,
         yes) | scan (nx, ny, dispaxis)

    if (l_statsec == "default" || l_statsec == "") {

        # if dispaxis is missing, assume GNIRS default (along columns)
        # if image extension is not 2-D, does something appropriate
        # if axis is shorter than needed, or algorithm becomes confused,
        #    do nothing (i.e., sect="")
        sect = ""
        if (dispaxis == 1 && nx/2 > 20)
            printf ("[%d:%d,*]\n", (nx/2-20), (nx/2+20)) | scan (sect)
        else if (ny/2 > 20)
            printf ("[*,%d:%d]\n", (ny/2-20), (ny/2+20)) | scan (sect)
        else
            printlog ("WARNING - NFNORMSLICES: could not determine section: \
                defaulting to [*,*]", l_logfile, verbose+)
    } else
        sect = l_statsec

    # Check/expand flat
    # This can be at a different shift to the dark/flexflat

    gemextn (l_flat, check = "exists,mef", process = "none", \
        index = "", extname = "", extversion = "", ikparams = "", \
        omit = "", replace = "", outfile = "dev$null", logfile="",
        glogpars="", verbose=l_verbose)
    if (gemextn.fail_count != 0 || gemextn.count != 1) {
        printlog ("ERROR - NFNORMSLICES: Bad or missing flat.", \
            l_logfile, verbose+) 
        goto clean
    }
    gemextn (l_flat, check = "exists,mef", process = "none", \
        index = "", extname = "", extversion = "", ikparams = "", \
        omit = "", replace = "", outfile = "STDOUT", \
        logfile="dev$null", glogpars="", verbose-) | scan (line)
    l_flat = line
    gemextn (l_flat, check = "", process = "expand", index = "", \
        extname = l_sci_ext, extversion = "1-", ikparams = "", \
        omit = "", replace = "", outfile = "dev$null", \
        logfile="dev$null", glogpars="", verbose-)
    if (nsci != gemextn.count) {
        printlog ("ERROR - NFNORMSLICES: Flats have different no \
            of extensions (" // nsci // ")", l_logfile, verbose+) 
        printlog ("                        to input flat (" \
            // gemextn.count // ")", l_logfile, verbose+) 
        goto clean
    }

    if (havedqvar) {
        gemextn (l_flat, proc="expand", check="", index="", \
            extname=l_var_ext, extversion="-", ikparams="", omit="", \
            replace="", outfile="dev$null", glogpars="", verbose-)
        if (no == (nsci == gemextn.count)) havedqvar = no
        gemextn (l_flat, proc="expand", check="", index="", \
            extname=l_dq_ext, extversion="-", ikparams="", omit="", \
            replace="", outfile="dev$null", glogpars="", verbose-)
        if (no == (nsci == gemextn.count)) havedqvar = no
        if (no == havedqvar) {
            printlog ("WARNING - NFNORMSLICES: no dq/variance data \
                in " // l_flat, l_logfile, verbose+) 
        }
    }

    # Check output not already present

    gemextn (l_output, check = "absent", process = "none", \
        index = "", extname = "", extversion = "", ikparams = "", \
        omit = "", replace = "", outfile = "dev$null", logfile="",
        glogpars="", verbose=l_verbose)
    if (gemextn.fail_count != 0 || gemextn.count != 1) {
        printlog ("ERROR - NFNORMSLICES: Bad or existing output.", \
            l_logfile, verbose+) 
        goto clean
    }
    gemextn (l_output, check = "absent", process = "none", \
        index = "", extname = "", extversion = "", ikparams = "", \
        omit = "", replace = "", outfile = "STDOUT", \
        logfile="dev$null", glogpars="", verbose-) | scan (line)
    l_output = line

    # Check each ifuflat for various preconditions

    scanfile = phulist
    while (fscan (scanfile, phu) != EOF) {

        keyfound = ""
        hselect(phu, "*PREPAR*", yes) | scan (keyfound)
        if (keyfound == "") {
            printlog ("ERROR - NFNORMSLICES: Image " // phu \
                // " not PREPAREd.", l_logfile, verbose+) 
            goto clean
        }

        keyfound = ""
        hselect(phu, "NSTRANSF", yes) | scan (keyfound)
        if (keyfound == "") {
            printlog ("ERROR - NFNORMSLICES: Image " // phu \
                // " not rectified in wavelength.", l_logfile, verbose+)
            goto clean
        }

        exptime = INDEF
        hselect (phu, l_key_exptime, yes) | scan (exptime)
        if (isindef (exptime)) {
            printlog ("WARNING - NFNORMSLICES: missing exposure time \
                in " // phu, l_logfile, verbose+) 
        } else if (no == isindef (prevexp)) {
            if (abs (exptime - prevexp) > 0.1 \
                && no == ("exposure" == strlwr (l_scale))) {

                printlog ("WARNING - NFNORMSLICES: exposure time \
                    changed", l_logfile, verbose+)
                printf ("                  %-28s: %6.2f\n", \
                    previmg, prevexp) | scan (line)
                printlog (line,  l_logfile, verbose+)
                printf ("                  %-28s: %6.2f\n", \
                    phu, exptime) | scan (line)
                printlog (line,  l_logfile, verbose+)
            }
        }
        previmg = phu; prevexp = exptime

    }

    # Combine IFU flat spectra

    if (n == 1) {

        gemextn ("@" // filelist, proc="none", check="exists", \
            index="", extname="", extversion="", ikparams="", omit="", \
            replace="", outfile="STDOUT", logfile="dev$null", \
            glogpars="", verbose=l_verbose) | scan (comboutptr)

        if (havedqvar) {
            gemextn (comboutptr, proc="expand", check="", index="", \
                extname=l_sci_ext, extversion="-", ikparams="", omit="", \
                replace="", outfile="dev$null", glogpars="", verbose-)
            if (0 == gemextn.count || 0 < gemextn.fail_count) {
                printlog ("ERROR - NFNORMSLICES: missing science data \
                    in " // comboutptr, l_logfile, verbose+) 
                goto clean
            }
            nextn = gemextn.count
            gemextn (comboutptr, proc="expand", check="", index="", \
                extname=l_var_ext, extversion="-", ikparams="", omit="", \
                replace="", outfile="dev$null", glogpars="", verbose-)
            if (no == (nextn == gemextn.count)) havedqvar = no
            gemextn (comboutptr, proc="expand", check="", index="", \
                extname=l_dq_ext, extversion="-", ikparams="", omit="", \
                replace="", outfile="dev$null", glogpars="", verbose-)
            if (no == (nextn == gemextn.count)) havedqvar = no
            if (no == havedqvar) {
                printlog ("WARNING - NFNORMSLICES: no dq/variance data \
                    in " // comboutptr, l_logfile, verbose+) 
            }
        }

    } else {

        printlog ("NFNORMSLICES: Combining spatial flats.", \
            l_logfile, l_verbose)

        gemcombine ("@" // filelist, combout, title="", \
            combine=l_combine, reject=l_reject, \
            masktype="goodvalue", maskvalue=0, \
            offsets="none", scale=l_scale, zero=l_zero, \
            weight=l_weight, statsec=sect, expname=l_key_exptime, \
            lthreshold=l_lthreshold, hthreshold=l_hthreshold, \
            nlow=l_nlow, nhigh=l_nhigh, nkeep=l_nkeep, mclip=l_mclip, \
            lsigma=l_lsigma, hsigma=l_hsigma, key_ron=l_key_ron,
            key_gain=l_key_gain, ron=l_ron, gain=l_gain, snoise=l_snoise, \
            sigscale=l_sigscale, pclip=l_pclip, grow=l_grow, bpmfile="", \
            nrejfile="", sci_ext=l_sci_ext, var_ext=l_var_ext, \
            dq_ext=l_dq_ext, fl_vardq+, logfile=l_logfile, \
            fl_dqprop+, verbose-)

        if (gemcombine.status != 0) {
            printlog ("ERROR - NFNORMSLICES: Error combining sky \
                flats.", l_logfile, verbose+)
            goto clean
        }

        comboutptr = combout

    }

    printlog ("NFNORMSLICES: Calculating relative slice levels.", \
        l_logfile, l_verbose)

    gemextn (comboutptr, check = "exists,mef", process = "expand", \
        index = "", extname = l_sci_ext, extversion = "1-", \
        ikparams = "", omit = "", replace = "", outfile = scilist, \
        logfile=l_logfile, glogpars="", verbose=l_verbose)
    if (gemextn.fail_count != 0 || gemextn.count == 0) {
        printlog ("ERROR - NFNORMSLICES: Missing internal data.", \
            l_logfile, verbose+) 
        goto clean
    }
    nver = gemextn.count

    gemextn (comboutptr, check="", process="expand", index="", \
        extname=l_var_ext, extversion="1-", ikparams="", \
        omit="", replace="", outfile=varlist, logfile=l_logfile,
        glogpars="", verbose=l_verbose)
    gemextn (comboutptr, check="", process="expand", index="", \
        extname=l_dq_ext, extversion="1-", ikparams="", \
        omit="", replace="", outfile=dqlist, logfile=l_logfile,
        glogpars="", verbose=l_verbose)

    # Output PHU

    imcopy (l_flat // ".fits[0]", l_output // ".fits[0,overwrite]", \
        >& "dev$null")

    # Output MDF

    extn = INDEF
    gemextn (comboutptr, check="exists,mef", process="expand", \
        index="0-", extname="MDF", extver="", ikparams="", \
        omit="name", replace="", outfile="STDOUT", \
        logfile="dev$null", glogpars="", verbose-) | scan (extn)
    if (no == isindef (extn))
        fxinsert (extn, l_output // ".fits[0]", groups="", verbose-,
            >& "dev$null")
    if (debug) gemextn (l_output)

    # List of source frames

    gemextn (l_flat, check="", process="expand", index="", \
        extname=l_sci_ext, extversion="1-", ikparams="", \
        omit="", replace="", outfile=ssrclist, logfile=l_logfile, \
        glogpars="", verbose=l_verbose)
    if (gemextn.fail_count != 0 || gemextn.count != nver) {
        printlog ("ERROR - NFNORMSLICES: Sky flats inconsistent with \
            lamp flats (cut before processing?).", l_logfile, verbose+) 
        goto clean
    }
    gemextn (l_flat, check="", process="expand", index="", \
        extname=l_var_ext, extversion="1-", ikparams="", \
        omit="", replace="", outfile=vsrclist, logfile=l_logfile,
        glogpars="", verbose=l_verbose)
    gemextn (l_flat, check="", process="expand", index="", \
        extname=l_dq_ext, extversion="1-", ikparams="", \
        omit="", replace="", outfile=dqsrclist, logfile=l_logfile,
        glogpars="", verbose=l_verbose)

    # List of output frames

    gemextn (l_output, check="", process="append", index="", \
        extname=l_sci_ext, extversion="1-" // nver, ikparams="append", \
        omit="", replace="", outfile=sdestlist, logfile=l_logfile, \
        glogpars="", verbose=l_verbose)
    gemextn (l_output, check="", process="append", index="", \
        extname=l_var_ext, extversion="1-" // nver, ikparams="append", \
        omit="", replace="", outfile=vdestlist, logfile=l_logfile,
        glogpars="", verbose=l_verbose)
    gemextn (l_output, check="", process="append", index="", \
        extname=l_dq_ext, extversion="1-" // nver, ikparams="append", \
        omit="", replace="", outfile=dqdestlist, logfile=l_logfile,
        glogpars="", verbose=l_verbose)

    joinlines (scilist // "," // varlist // "," // dqlist // "," \
        // ssrclist // "," // sdestlist // "," \
        // vsrclist // "," // vdestlist // "," // dqsrclist // "," \
        // dqdestlist, "", output=comblist, delim=" ", \
        missing="none", maxchars=1000, shortest-, verbose-)

    # Processing and normalisation in two steps to allow for normzn by
    # middle order

    scanfile = comblist
    i = 0
    n = nver / 2 + 1

    while (fscan (scanfile,img,varimg,dqimg,ssrc,sdest,vsrc,vdest,
                  dqsrc,dqdest) != EOF) {
    i = i + 1

        if (debug) {
            print ("img:      " // img)
            print ("ssrc:     " // ssrc)
            print ("sdest:    " // sdest)
        }

        if (debug)
            imstat (img, fields="image,npix,mean,stddev,min,max,midpt")

        tnx = -1; tny = -1
        hselect (img, "i_naxis1,i_naxis2", yes) | scan (tnx, tny)

        if ((dispaxis == 1 && tnx != nx) || (dispaxis != 1 && tny != ny)) {
            printlog ("ERROR - NFNORMSLICES: input slice lengths don't match",
                      l_logfile, verbose+) 
            goto clean
        }

        # Mask unilluminated regions for imstat:
        imdelete (tmpimg, verify-, >& "dev$null")
        imexpr("b & 249 ? -99999. : a", tmpimg, a=img, b=dqimg, verbose-)

        imstat (tmpimg//sect, fields="midpt,stddev", lower=0., upper=INDEF, \
            nclip=0, lsigma=3, usigma=3, binwidth=0.1, format-,
            cache-) | scan (mddata,sigma)
        if (debug) {
            print ("1: sect: " // sect)
            print ("   mddata: " // mddata)
            print ("   sigma:  " // sigma)
        }

        imstat (tmpimg//sect, fields="midpt", \
            lower=max(0.3*mddata,(mddata-3*sigma)),upper=(mddata+3*sigma), \
            nclip=0, lsigma=3, usigma=3, binwidth=0.1, format-,
            cache-) | scan (mddata)
        if (debug) {
            print ("2: mddata: " // mddata)
        }

        print (i // " " // ssrc // " " // sdest // " " // mddata \
            // " " // vsrc // " " // vdest // " " // dqsrc // " " // dqdest, \
            >> comblist2)

        if (i == n) oldmddata = mddata
    }

    # So now we do the normlzn and assemble the output

    printlog ("NFNORMSLICES: Normalizing response.", l_logfile, l_verbose)

    scanfile = comblist2

    while (fscan (scanfile, i, ssrc, sdest, mddata, \
        vsrc, vdest, dqsrc, dqdest) != EOF) {

        if (debug) {
            print ("ssrc:     " // ssrc)
            print ("sdest:    " // sdest)
            print ("mddata:   " // mddata)
        }

        correcn = mddata / oldmddata  # divide data by this to remove variation

        if (debug) {
            print ("correction: " // correcn)
            # print(correcn, >>l_output)
            print ("scaling flat")
        }
        imdelete (tmpimg, verify-, >& "dev$null")
        imexpr("b & 8 ? 0. : a * c", tmpimg, a=ssrc, b=dqsrc, c=correcn,
               refim="a", outtype="ref", verbose-)
        imcopy (tmpimg, sdest, >& "dev$null")

        if (havedqvar) {
            if (debug) print ("scaling var")
            imdelete (tmpimg, verify-, >& "dev$null")
            imexpr ("b & 8 ? 0. : a * c*c", tmpimg, a=vsrc, b=dqsrc, c=correcn,
                    refim="a", outtype="ref", verbose-)
            imcopy (tmpimg, vdest, >& "dev$null")

            if (debug) print ("copying dq")
            imcopy (dqsrc, dqdest, >& "dev$null")
        }

    }

    gemdate () 
    if (debug) print (l_output)
    gemhedit (l_output // "[0]", "NFNORMSL", gemdate.outdate, 
        "UT Time stamp for NFNORMSLICES", delete-)
    gemhedit (l_output // "[0]", "GEM-TLM", gemdate.outdate, 
        "UT Last modification with GEMINI", delete-)
    if (l_title != "") {
        gemhedit (l_output // "[0]", "i_title", l_title, "", delete-) 
    }

    status = 0


clean:

    if (debug) print ("clean")
    delete (filelist, verify-, >& "dev$null")
    delete (phulist, verify-, >& "dev$null")
    delete (scilist, verify-, >& "dev$null")
    delete (varlist, verify-, >& "dev$null")
    delete (dqlist, verify-, >& "dev$null")
    delete (ssrclist, verify-, >& "dev$null")
    delete (sdestlist, verify-, >& "dev$null")
    delete (vsrclist, verify-, >& "dev$null")
    delete (vdestlist, verify-, >& "dev$null")
    delete (dqsrclist, verify-, >& "dev$null")
    delete (dqdestlist, verify-, >& "dev$null")
    delete (comblist, verify-, >& "dev$null")
    delete (comblist2, verify-, >& "dev$null")

    imdelete (combout, verify-, >& "dev$null")
    imdelete (tmpimg, verify-, >& "dev$null")

    printlog (" ", l_logfile, l_verbose) 
    printlog ("NFNORMSLICES done", l_logfile, l_verbose) 
    printlog ("------------------------------------------------------\
        --------------------------", l_logfile, l_verbose)
        
end
